@@ -1,8 +1,6 @@
-from __future__ import absolute_import
-from __future__ import print_function
+# coding: utf-8
 
-__all__ = ['BaseConstructor', 'SafeConstructor', 'Constructor',
-           'ConstructorError', 'RoundTripConstructor']
+from __future__ import print_function, absolute_import, division
 
 import collections
 import datetime
@@ -12,20 +10,18 @@ import re
 import sys
 import types
 
-try:
-    from .error import *
-    from .nodes import *
-    from .compat import (utf8, builtins_module, to_str, PY2, PY3, ordereddict,
-                         text_type)
-    from .comments import *
-    from .scalarstring import *
-except (ImportError, ValueError):  # for Jython
-    from ruamel.yaml.error import *
-    from ruamel.yaml.nodes import *
-    from ruamel.yaml.compat import (utf8, builtins_module, to_str, PY2, PY3, 
-                                    ordereddict, text_type)
-    from ruamel.yaml.comments import *
-    from ruamel.yaml.scalarstring import *
+from typing import Any, Dict  # NOQA
+
+from ruamel.yaml.error import *                               # NOQA
+from ruamel.yaml.nodes import *                               # NOQA
+from ruamel.yaml.compat import (utf8, builtins_module, to_str, PY2, PY3,
+                                ordereddict, text_type)
+from ruamel.yaml.comments import *                               # NOQA
+from ruamel.yaml.scalarstring import *                           # NOQA
+from ruamel.yaml.timestamp import TimeStamp
+
+__all__ = ['BaseConstructor', 'SafeConstructor', 'Constructor',
+           'ConstructorError', 'RoundTripConstructor']
 
 
 class ConstructorError(MarkedYAMLError):
@@ -34,14 +30,15 @@ class ConstructorError(MarkedYAMLError):
 
 class BaseConstructor(object):
 
-    yaml_constructors = {}
-    yaml_multi_constructors = {}
+    yaml_constructors = {}   # type: Dict[Any, Any]
+    yaml_multi_constructors = {}   # type: Dict[Any, Any]
 
-    def __init__(self):
+    def __init__(self, preserve_quotes=None):
         self.constructed_objects = {}
         self.recursive_objects = {}
         self.state_generators = []
         self.deep_construct = False
+        self._preserve_quotes = preserve_quotes
 
     def check_data(self):
         # If there are more documents available?
@@ -73,6 +70,9 @@ class BaseConstructor(object):
         return data
 
     def construct_object(self, node, deep=False):
+        """deep is True when creating an object/mapping recursively,
+        in that case want the underlying elements available during construction
+        """
         if node in self.constructed_objects:
             return self.constructed_objects[node]
         if deep:
@@ -132,6 +132,9 @@ class BaseConstructor(object):
         return node.value
 
     def construct_sequence(self, node, deep=False):
+        """deep is True when creating an object/mapping recursively,
+        in that case want the underlying elements available during construction
+        """
         if not isinstance(node, SequenceNode):
             raise ConstructorError(
                 None, None,
@@ -141,6 +144,9 @@ class BaseConstructor(object):
                 for child in node.value]
 
     def construct_mapping(self, node, deep=False):
+        """deep is True when creating an object/mapping recursively,
+        in that case want the underlying elements available during construction
+        """
         if not isinstance(node, MappingNode):
             raise ConstructorError(
                 None, None,
@@ -187,12 +193,14 @@ class BaseConstructor(object):
 
     @classmethod
     def add_constructor(cls, tag, constructor):
+        # type: (Any, Any) -> None
         if 'yaml_constructors' not in cls.__dict__:
             cls.yaml_constructors = cls.yaml_constructors.copy()
         cls.yaml_constructors[tag] = constructor
 
     @classmethod
     def add_multi_constructor(cls, tag_prefix, multi_constructor):
+        # type: (Any, Any) -> None
         if 'yaml_multi_constructors' not in cls.__dict__:
             cls.yaml_multi_constructors = cls.yaml_multi_constructors.copy()
         cls.yaml_multi_constructors[tag_prefix] = multi_constructor
@@ -250,6 +258,9 @@ class SafeConstructor(BaseConstructor):
             node.value = merge + node.value
 
     def construct_mapping(self, node, deep=False):
+        """deep is True when creating an object/mapping recursively,
+        in that case want the underlying elements available during construction
+        """
         if isinstance(node, MappingNode):
             self.flatten_mapping(node)
         return BaseConstructor.construct_mapping(self, node, deep=deep)
@@ -288,9 +299,9 @@ class SafeConstructor(BaseConstructor):
             return sign*int(value[2:], 16)
         elif value.startswith('0o'):
             return sign*int(value[2:], 8)
-        elif value[0] == '0':
+        elif self.processing_version != (1, 2) and value[0] == '0':
             return sign*int(value, 8)
-        elif ':' in value:
+        elif self.processing_version != (1, 2) and ':' in value:
             digits = [int(part) for part in value.split(':')]
             digits.reverse()
             base = 1
@@ -363,7 +374,7 @@ class SafeConstructor(BaseConstructor):
         u'''^(?P<year>[0-9][0-9][0-9][0-9])
           -(?P<month>[0-9][0-9]?)
           -(?P<day>[0-9][0-9]?)
-          (?:(?:[Tt]|[ \\t]+)
+          (?:((?P<t>[Tt])|[ \\t]+)   # explictly not retaining extra spaces
           (?P<hour>[0-9][0-9]?)
           :(?P<minute>[0-9][0-9])
           :(?P<second>[0-9][0-9])
@@ -371,10 +382,10 @@ class SafeConstructor(BaseConstructor):
           (?:[ \\t]*(?P<tz>Z|(?P<tz_sign>[-+])(?P<tz_hour>[0-9][0-9]?)
           (?::(?P<tz_minute>[0-9][0-9]))?))?)?$''', re.X)
 
-    def construct_yaml_timestamp(self, node):
-        value = self.construct_scalar(node)
-        match = self.timestamp_regexp.match(node.value)
-        values = match.groupdict()
+    def construct_yaml_timestamp(self, node, values=None):
+        if values is None:
+            match = self.timestamp_regexp.match(node.value)
+            values = match.groupdict()
         year = int(values['year'])
         month = int(values['month'])
         day = int(values['day'])
@@ -389,6 +400,8 @@ class SafeConstructor(BaseConstructor):
             while len(fraction) < 6:
                 fraction += '0'
             fraction = int(fraction)
+            if len(values['fraction']) > 6 and int(values['fraction'][6]) > 4:
+                fraction += 1
         delta = None
         if values['tz_sign']:
             tz_hour = int(values['tz_hour'])
@@ -816,6 +829,11 @@ class RoundTripConstructor(SafeConstructor):
 
         if node.style == '|' and isinstance(node.value, text_type):
             return PreservedScalarString(node.value)
+        elif self._preserve_quotes and isinstance(node.value, text_type):
+            if node.style == "'":
+                return SingleQuotedScalarString(node.value)
+            if node.style == '"':
+                return DoubleQuotedScalarString(node.value)
         return node.value
 
     def construct_yaml_str(self, node):
@@ -873,7 +891,7 @@ class RoundTripConstructor(SafeConstructor):
                 value = self.construct_object(value_node, deep=False)
             return value
 
-        #merge = []
+        # merge = []
         merge_map_list = []
         index = 0
         while index < len(node.value):
@@ -883,10 +901,10 @@ class RoundTripConstructor(SafeConstructor):
                 if isinstance(value_node, MappingNode):
                     merge_map_list.append(
                         (index, constructed(value_node)))
-                    #self.flatten_mapping(value_node)
-                    #merge.extend(value_node.value)
+                    # self.flatten_mapping(value_node)
+                    # merge.extend(value_node.value)
                 elif isinstance(value_node, SequenceNode):
-                    #submerge = []
+                    # submerge = []
                     for subnode in value_node.value:
                         if not isinstance(subnode, MappingNode):
                             raise ConstructorError(
@@ -896,11 +914,11 @@ class RoundTripConstructor(SafeConstructor):
                                 % subnode.id, subnode.start_mark)
                         merge_map_list.append(
                             (index, constructed(subnode)))
-                    #    self.flatten_mapping(subnode)
-                    #    submerge.append(subnode.value)
-                    #submerge.reverse()
-                    #for value in submerge:
-                    #    merge.extend(value)
+                    #     self.flatten_mapping(subnode)
+                    #     submerge.append(subnode.value)
+                    # submerge.reverse()
+                    # for value in submerge:
+                    #     merge.extend(value)
                 else:
                     raise ConstructorError(
                         "while constructing a mapping", node.start_mark,
@@ -912,10 +930,12 @@ class RoundTripConstructor(SafeConstructor):
                 index += 1
             else:
                 index += 1
-        #print ('merge_map_list', merge_map_list)
         return merge_map_list
-        #if merge:
-        #    node.value = merge + node.value
+        # if merge:
+        #     node.value = merge + node.value
+
+    def _sentinel(self):
+        pass
 
     def construct_mapping(self, node, maptyp, deep=False):
         if not isinstance(node, MappingNode):
@@ -923,10 +943,7 @@ class RoundTripConstructor(SafeConstructor):
                 None, None,
                 "expected a mapping node, but found %s" % node.id,
                 node.start_mark)
-        if isinstance(node, MappingNode):
-            merge_map = self.flatten_mapping(node)
-            if merge_map:
-                maptyp.add_yaml_merge(merge_map)
+        merge_map = self.flatten_mapping(node)
         # mapping = {}
         if node.comment:
             maptyp._yaml_add_comment(node.comment[:2])
@@ -936,13 +953,19 @@ class RoundTripConstructor(SafeConstructor):
             from ruamel.yaml.serializer import templated_id
             if not templated_id(node.anchor):
                 maptyp.yaml_set_anchor(node.anchor)
+        last_key, last_value = None, self._sentinel
         for key_node, value_node in node.value:
             # keys can be list -> deep
             key = self.construct_object(key_node, deep=True)
             # lists are not hashable, but tuples are
             if not isinstance(key, collections.Hashable):
                 if isinstance(key, list):
-                    key = tuple(key)
+                    key = CommentedKeySeq(key)
+                    if key_node.flow_style is True:
+                        key.fa.set_flow_style()
+                    elif key_node.flow_style is False:
+                        key.fa.set_block_style()
+                    # key = tuple(key)
             if PY2:
                 try:
                     hash(key)
@@ -957,14 +980,28 @@ class RoundTripConstructor(SafeConstructor):
                         "while constructing a mapping", node.start_mark,
                         "found unhashable key", key_node.start_mark)
             value = self.construct_object(value_node, deep=deep)
+            if key_node.comment and len(key_node.comment) > 4 and \
+               key_node.comment[4]:
+                if last_value is None:
+                    key_node.comment[0] = key_node.comment.pop(4)
+                    maptyp._yaml_add_comment(key_node.comment, value=last_key)
+                else:
+                    key_node.comment[2] = key_node.comment.pop(4)
+                    maptyp._yaml_add_comment(key_node.comment, key=key)
+                key_node.comment = None
             if key_node.comment:
                 maptyp._yaml_add_comment(key_node.comment, key=key)
             if value_node.comment:
                 maptyp._yaml_add_comment(value_node.comment, value=key)
             maptyp._yaml_set_kv_line_col(
                 key, [key_node.start_mark.line, key_node.start_mark.column,
-                value_node.start_mark.line, value_node.start_mark.column])
+                      value_node.start_mark.line, value_node.start_mark.column])
             maptyp[key] = value
+            last_key, last_value = key, value  # could use indexing
+        # do this last, or <<: before a key will prevent insertion in instances
+        # of collections.OrderedDict (as they have no __contains__
+        if merge_map:
+            maptyp.add_yaml_merge(merge_map)
 
     def construct_setting(self, node, typ, deep=False):
         if not isinstance(node, MappingNode):
@@ -1000,7 +1037,7 @@ class RoundTripConstructor(SafeConstructor):
                     raise ConstructorError(
                         "while constructing a mapping", node.start_mark,
                         "found unhashable key", key_node.start_mark)
-            value = self.construct_object(value_node, deep=deep)
+            value = self.construct_object(value_node, deep=deep)  # NOQA
             if key_node.comment:
                 typ._yaml_add_comment(key_node.comment, key=key)
             if value_node.comment:
@@ -1077,6 +1114,75 @@ class RoundTripConstructor(SafeConstructor):
         yield data
         self.construct_setting(node, data)
 
+    def construct_undefined(self, node):
+        try:
+            data = CommentedMap()
+            data._yaml_set_line_col(node.start_mark.line, node.start_mark.column)
+            if node.flow_style is True:
+                data.fa.set_flow_style()
+            elif node.flow_style is False:
+                data.fa.set_block_style()
+            data.yaml_set_tag(node.tag)
+            yield data
+            self.construct_mapping(node, data)
+        except:
+            raise ConstructorError(
+                None, None,
+                "could not determine a constructor for the tag %r" %
+                utf8(node.tag),
+                node.start_mark)
+
+    def construct_yaml_timestamp(self, node):
+        match = self.timestamp_regexp.match(node.value)
+        values = match.groupdict()
+        if not values['hour']:
+            return SafeConstructor.construct_yaml_timestamp(self, node, values)
+        for part in ['t', 'tz_sign', 'tz_hour', 'tz_minute']:
+            if values[part]:
+                break
+        else:
+            return SafeConstructor.construct_yaml_timestamp(self, node, values)
+        year = int(values['year'])
+        month = int(values['month'])
+        day = int(values['day'])
+        hour = int(values['hour'])
+        minute = int(values['minute'])
+        second = int(values['second'])
+        fraction = 0
+        if values['fraction']:
+            fraction = values['fraction'][:6]
+            while len(fraction) < 6:
+                fraction += '0'
+            fraction = int(fraction)
+            if len(values['fraction']) > 6 and int(values['fraction'][6]) > 4:
+                fraction += 1
+        delta = None
+        if values['tz_sign']:
+            tz_hour = int(values['tz_hour'])
+            tz_minute = int(values['tz_minute'] or 0)
+            delta = datetime.timedelta(hours=tz_hour, minutes=tz_minute)
+            if values['tz_sign'] == '-':
+                delta = -delta
+        if delta:
+            dt = datetime.datetime(year, month, day, hour, minute)
+            dt -= delta
+            data = TimeStamp(dt.year, dt.month, dt.day, dt.hour, dt.minute,
+                             second, fraction)
+            data._yaml['delta'] = delta
+            tz = values['tz_sign'] + values['tz_hour']
+            if values['tz_minute']:
+                tz += ':' + values['tz_minute']
+            data._yaml['tz'] = tz
+        else:
+            data = TimeStamp(year, month, day, hour, minute, second, fraction)
+            if values['tz']:  # no delta
+                data._yaml['tz'] = values['tz']
+
+        if values['t']:
+            data._yaml['t'] = True
+        return data
+
+
 RoundTripConstructor.add_constructor(
     u'tag:yaml.org,2002:null',
     RoundTripConstructor.construct_yaml_null)
